# IssueTrackerProduct
#
# Peter Bengtsson <mail@peterbe.com>
# Date: 06/05 -02
#

CHANGELOG="""
   - 0.4.3     Started making the ZClass product a python product
"""
__version__='0.4.3a'
__doc__="""IssueTrackerProduct is the easiest bug/issue tracker
system to use for Zope.
By Peter Bengtsson <mail@peterbe.com>"""

# python
import string, os

# Zope
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Globals import  MessageDialog, Persistent, InitializeClass, package_home,\
                     DTMLFile
from Products.PythonScripts.standard import html_quote
from Products.PythonScripts.PythonScript import PythonScript
from OFS import SimpleItem, ObjectManager, Folder, PropertySheets
from DocumentTemplate import sequence
from AccessControl import ClassSecurityInfo
from Products.ZCatalog.CatalogAwareness import CatalogAware
from zLOG import LOG, ERROR, INFO, PROBLEM

# Product
import IssueTrackerUtils
from IssueTrackerUtils import cookIdAndTitle,  \
                              DateStr2DateDate, ValidEmailAddress,\
                              getNowdate, GenerateNotifyableId, \
                              getRandomString, fixDictofLists, \
                              getLaterDate, AddParam2URL, getFromstring, \
                              ss

# constants
ICON_LOCATION = 'misc_/IssueTrackerProduct'

# properties
DEFAULT_TYPES = ('general','bug report','missing page','feature request',\
                 'content request','usability','other')
DEFAULT_TYPE = DEFAULT_TYPES[0]
DEFAULT_URGENCIES = ('low','normal','high','critical')
DEFAULT_URGENCY = DEFAULT_URGENCIES[1]
DEFAULT_SECTIONS_OPTIONS = ('IssueTracker','General','Homepage','Other')
DEFAULT_SECTIONS = [DEFAULT_SECTIONS_OPTIONS[1]]
DEFAULT_WHEN_IGNORE_WORD = 'ignored'
DEFAULT_DISPLAY_DATE = '%d/%m %Y %H:%M'
DEFAULT_SITEMASTER_NAME = 'Issue Tracker Master'
DEFAULT_SITEMASTER_EMAIL = 'issuetracker@localhost'
DEFAULT_ODD_BGCOLOR = 'silver'
DEFAULT_EVEN_BGCOLOR = 'gainsboro'
DEFAULT_DEFAULT_BATCH_SIZE = 25
DEFAULT_ISSUEPREFIX = ''
DEFAULT_STATUSES = ('open','taken','on hold','rejected','completed')
DEFAULT_NAME_COOKIEKEY = '__issuetracker_fromname'
DEFAULT_EMAIL_COOKIEKEY = '__issuetracker_email'
DEFAULT_DISPLAY_FORMAT_COOKIEKEY = '__issuetracker_display_format'
DEFAULT_DISPLAY_FORMATS = ('plaintext','structuredtext')
DEFAULT_DEFAULT_DISPLAY_FORMAT = DEFAULT_DISPLAY_FORMATS[0]
DEFAULT_RANDOMID_LENGTH = 4
DEFAULT_STATUSES_TRANSLATION_VERBS = ('open|open','take|take',\
                                      'put on hold|put on hold',\
                                      'reject|reject','complete|complete')
DEFAULT_STATUSES_TRANSLATION_ADJECTIVES =\
                    ('open|open','taken|taken','on hold|on hold',\
                     'rejected|rejected','completed|completed')

NATIVE_PROPERTIES = ('title','types','urgencies','sections_options', 'defaultsections',
                'when_ignore_word','display_date','sitemaster_name',
                'sitemaster_email','odd_bgcolor','even_bgcolor','default_type',
                'default_urgency','default_batch_size','issueprefix',
                'statuses','notify_groups',
                'name_cookiekey','email_cookiekey','display_format_cookiekey',
                'display_formats','default_display_format','randomid_length',
                'statuses_translation_verbs','statuses_translation_adjectives')

# Language constants
FAILURE_SUBJECT = 'Subject'
FAILURE_SUBJECT_DESC = 'You must at least enter a subject' 
NONAME_NOEMAIL = 'No name or email'
NAME_EMAIL_HIDDEN = 'Name and email hidden'
SORT_BY = 'Click to sort by'
SORT_REVERSE = 'Click one more time and the sorting will be reversed'

manage_addIssueTrackerForm = PageTemplateFile('zpt/addIssueTrackerForm', globals())

def manage_addIssueTracker(dispatcher, id, title='', REQUEST=None):
    """ add IssueTracker instance via the web """
    dest = dispatcher.Destination()
    issuetracker = IssueTracker(id, title)
    dest._setObject(id, issuetracker)
    self = dest._getOb(id)
    
    
    self.DeployStandards()
    self.InitZCatalog()

    if REQUEST is not None:
        # whereto next?
        if REQUEST.has_key('addandedit'):
            REQUEST.RESPONSE.redirect(self.absolute_url()+'/manage_workspace')
        elif REQUEST.has_key('DestinationURL'):
            REQUEST.RESPONSE.redirect(REQUEST.DestinationURL+'/manage_workspace')
        else:
            REQUEST.RESPONSE.redirect(REQUEST.URL1+'/manage_workspace')
    
    

class IssueTracker(Folder.Folder, CatalogAware, Persistent              
                   ):
    """ IssueTracker class """

    meta_type = 'Issue Tracker'

    security=ClassSecurityInfo()

    manage_options = Folder.Folder.manage_options[:2] + \
        ({'label':'Properties', 'action':'manage_editIssueTrackerPropertiesForm'},
         {'label':'Management', 'action':'manage_ManagementForm'}) \
        + Folder.Folder.manage_options[3:]
    
    
    def __init__(self, id, title='',
                 sitemaster_name=DEFAULT_SITEMASTER_NAME,
                 sitemaster_email=DEFAULT_SITEMASTER_EMAIL):
        """ Init IssueTracker class """
        self.id = str(id)
        self.title = str(title)
        
        self.types              = DEFAULT_TYPES
        self.urgencies          = DEFAULT_URGENCIES
        self.sections_options   = DEFAULT_SECTIONS_OPTIONS
        self.defaultsections    = DEFAULT_SECTIONS
        self.when_ignore_word   = DEFAULT_WHEN_IGNORE_WORD
        self.display_date       = DEFAULT_DISPLAY_DATE
        self.sitemaster_name    = sitemaster_name
        self.sitemaster_email   = sitemaster_email
        self.odd_bgcolor        = DEFAULT_ODD_BGCOLOR
        self.even_bgcolor       = DEFAULT_EVEN_BGCOLOR
        self.default_type       = DEFAULT_TYPE
        self.default_urgency    = DEFAULT_URGENCY
        self.default_batch_size = DEFAULT_DEFAULT_BATCH_SIZE
        self.issueprefix        = DEFAULT_ISSUEPREFIX
        self.statuses           = DEFAULT_STATUSES
        self.name_cookiekey     = DEFAULT_NAME_COOKIEKEY
        self.email_cookiekey    = DEFAULT_EMAIL_COOKIEKEY
        self.display_format_cookiekey = DEFAULT_DISPLAY_FORMAT_COOKIEKEY
        self.display_formats    = DEFAULT_DISPLAY_FORMATS
        self.default_display_format = DEFAULT_DEFAULT_DISPLAY_FORMAT
        self.randomid_length = DEFAULT_RANDOMID_LENGTH
        self.statuses_translation_verbs = DEFAULT_STATUSES_TRANSLATION_VERBS
        self.statuses_translation_adjectives = \
                    DEFAULT_STATUSES_TRANSLATION_ADJECTIVES
        self.notify_groups = []



    # forms
    manage_editIssueTrackerPropertiesForm = \
                    DTMLFile('dtml/editIssueTrackerPropertiesForm',\
                             globals(), native_properties=NATIVE_PROPERTIES)
    manage_ManagementForm = DTMLFile('dtml/ManagementForm', globals())

    # scripts
    TranslateStatus = IssueTrackerUtils.TranslateStatus
    timeSince = IssueTrackerUtils.timeSince
    def ShowDescription(self, text, display_format=''):
        """ pass on to utilities module """
        return IssueTrackerUtils.ShowDescription(text, display_format)


    ## This is "under construction"
    def standard_error_message(self, REQUEST, RESPONSE=None):
        """ custom error message """
        return getattr(self, 'standard_error_message.html')
                       

    def manage_editIssueTrackerProperties(self, REQUEST):
        """ save all IssueTracker related issues """
        hk = self.REQUEST.has_key
        get = self.REQUEST.get
        strings = ['title','when_ignore_word','display_date','sitemaster_name',
                   'sitemaster_email','odd_bgcolor','even_bgcolor',
                   'default_type','default_urgency','issueprefix',
                   'name_cookiekey','email_cookiekey','display_format_cookiekey',
                   'default_display_format']
        lists = ['types','urgencies','sections_options','defaultsections',
                 'statuses','display_formats','statuses_translation_verbs',
                 'statuses_translation_adjectives']
        ints = ['default_batch_size','randomid_length']
        for each in strings:
            if hk(each) and type(get(each))==type('s'):
                exec 'self.%s = "%s"' % (each, get(each).strip())
        for each in ints:
            if hk(each):
                if type(get(each))==type(2):
                    exec "self.%s = %s" % (each, get(each))
                else:
                    LOG(self.meta_type, WARNING,
                        '%s not integer'%get(each), '')
        for each in lists:
            if hk(each) and type(get(each))==type([]):
                dummy = []
                for item in get(each):
                    if item.strip() != '':
                        dummy.append(item.strip())
                exec "self.%s = %s" % (each, dummy)
                
        # for the custom properties
        self.manage_editProperties(REQUEST)
        return self.manage_editIssueTrackerPropertiesForm(REQUEST,
                        manage_tabs_message='IssueTracker properties updated.')

    def DeployStandards(self, remove_oldstuff=0, DestinationURL=None,
                        initzcatalog=0):
        """ copy images and other documents into the instance unless they
            are already there
        """

        t={}
        
        if initzcatalog:
            self.InitZCatalog()
            
            
        # create folders
        root = self.getRoot()
        for f in ['actionbuttons', 'mainbuttons', 'notifyables', 'www']:
            if not hasattr(root, f):
                root.manage_addFolder(f)
                t[f]='Folder'

        osj = os.path.join
        standards_home = osj(package_home(globals()),'standards')
        t = self._deployImages(root, standards_home, t=t)

        mainbuttons_home = osj(standards_home,'mainbuttons')
        t = self._deployImages(root.mainbuttons, mainbuttons_home, t=t)

        actionbuttons_home = osj(standards_home,'actionbuttons')
        t = self._deployImages(root.actionbuttons, actionbuttons_home, t=t)

        www_home = osj(standards_home,'www')
        t = self._deployImages(root.www, www_home, t=t)

        # shortcut
        addPG = root.manage_addProduct['PageTemplates'].manage_addPageTemplate
        
        for filestr in os.listdir(standards_home):
            if filestr[-5:] == '.dtml':
                id, title = cookIdAndTitle(filestr.replace('.dtml',''))
                if not hasattr(root, id):
                    file = DTMLFile('standards/%s'%filestr.replace('.dtml',''), globals()).read()
                    root.manage_addDTMLDocument(id, title, file=file)
                    t[id] ="DTML Document"
            elif filestr[-4:] == '.zpt':
                id, title = cookIdAndTitle(filestr.replace('.zpt',''))
                if not hasattr(root, id):
                    file = open(osj(standards_home,filestr)).read()
                    addPG(id, title=title, text=file)
                    t[id]="Page Template"                    
            elif filestr[-3:] == '.py':
                id, title = cookIdAndTitle(filestr.replace('.py',''))
                if hasattr(root, id) and remove_oldstuff:
                    root.manage_delObjects([id])
                if not hasattr(root, id):
                    file = open(osj(standards_home, filestr)).read()
                    id = root._setObject(id, PythonScript(id))
                    root._getOb(id).write(file)
                    t[id]="Script (Python)"
        if DestinationURL:
            msg = ''
            for k,v in t.items():
                msg = "%s (%s) <br>%s"%(k,v,msg)
            url = AddParam2URL(DestinationURL,{'manage_tabs_message':\
                                "Standard objects deployed<br>%s"%msg})
            self.REQUEST.RESPONSE.redirect(url)
        else:
            return "Standard objects deployed\n%s"%t

    def _deployImages(self, destination, dir, extension='.gif', t={}):
        """ do the actual deployment of images in a dir """
        osj = os.path.join
        for filestr in os.listdir(dir):
            if filestr.find(extension) > -1:
                # take the image
                id, title = cookIdAndTitle(filestr)
                if not hasattr(destination, id):
                    destination.manage_addImage(id, title=title, \
                          file=open(osj(dir, filestr),'rb').read())
                    t[id]="Image"
        return t

                
    def InitZCatalog(self):
        """ create a ZCatalog called 'Catalog' and change its properties
        accordingly """
        if not 'Catalog' in self.objectIds('ZCatalog'):
            self.manage_addProduct['ZCatalog'].manage_addZCatalog('Catalog','')
        zcatalog = self.Catalog
        indexes = zcatalog._catalog.indexes
        
        #shortcut
        pluginindexes = zcatalog.Indexes.manage_addProduct['PluginIndexes']
        
        add=('status','description','comment','fromname','email','url2issue')
        for adder in add:
            if not indexes.has_key(adder):
                pluginindexes.manage_addTextIndex(adder)


    def getRoot(self):
        """ Get the root instance object """
        if self.meta_type == 'Issue Tracker':
            return self
        else:
            for each in self.superValues('Issue Tracker'):
                if each.id in self.REQUEST.URL1.split('/'):
                    return each

    def getRootURL(self):
        """ quick wrapper around getRoot() """
        return self.getRoot().absolute_url()
        
    def _fillRequest(self, dict):
        """ add properties as empty if not existing """
        propertiesNulls	= ['title','fromname','email','description',
                           'url2issue','urgency','type','display_format']
        propertiesZero 	= ['confidential','hide_me']
        propertiesEmpty	= ['sections']
        
        nullified = []
        for property in propertiesNulls:
            if not dict.has_key(property):
                dict[property] =''
                
        for property in propertiesZero:
            if not dict.has_key(property):
                dict[property] = 0
                
        for property in propertiesEmpty:
            if not dict.has_key(property):
                dict[property]=[]
        
        # Also, for all list variables, make sure that only
        # one item doesn't make it a string
        for property in propertiesEmpty:
            if type(dict[property]) == type(''):
                dict[property] = [dict[property]]
                

        # special one
        try:
            if dict['url2issue'][0] != '/' and dict['url2issue'][:4] !='http':
                dict['url2issue'] = 'http://%s'%dict['url2issue']
        except:pass # no big deal

        return dict

    
    def SubmitIssue(self):
        """ This is the method to create an Issue Tracker Issue. It
        relies only on the REQUEST object.
        1) Check data
        2) Try to create issue
            2a) If success, RESPONSE.redirect to issue plus Thank you message
            2b) If failure, print failed data and urge to submit again
        """

        request = self.REQUEST

        failures = {}
        has_cookie = self.has_cookie
        get_cookie = self.get_cookie
        set_cookie = self.set_cookie

        # check for existance, by adding all missing properties
        # as empty instead of non-existant.
        request.set('title',request['subject'])
        request = self._fillRequest(request)
        
        #
        # Tune the data a bit
        #

        # strip whitespace
        for property in ['title','fromname','email','url2issue','display_format']:
            request[property] = request[property].strip()
        # Special treatment needed in case STX is used upon display
        request['description'] = request['description'].strip()+' '

        # use cookie if not else specified
        if request.email=='' and has_cookie(self.email_cookiekey):
            request['email'] = get_cookie(self.email_cookiekey)
        if request.fromname=='' and has_cookie(self.name_cookiekey):
            request['fromname'] = get_cookie(self.name_cookiekey)
        if request.display_format=='' and has_cookie(self.display_format_cookiekey):
            request['display_format'] = get_cookie(self.display_format_cookiekey)

        # append the default sections if not specified
        if len(request.sections) == 0:
            request.sections = self.defaultsections

        #
        # Check data
        #

        if request.subject == '':
            failures[FAILURE_SUBJECT] = FAILURE_SUBJECT_DESC

        valid_emailaddress = 1
        # to prevent problems with sending mail
        if not ValidEmailAddress(request.email):
            valid_emailaddress = 0

        if len(failures) >0:
            return self.ExplainSubmitErrors(failures)

        #
        # Let's submit the issue!
        #

        prefix = self.issueprefix
        genid = self._generateID("Issue", self.randomid_length, prefix)

        # if these are valid, save them
        if request.fromname <> '':
            set_cookie(self.name_cookiekey, request.fromname)
        if request.display_format in self.display_formats:
            set_cookie(self.display_format_cookiekey,request.display_format)
        if valid_emailaddress:
            set_cookie(self.email_cookiekey, request.email)


        # filter out empty item from sections
        sections_newlist=[]
        for section in request.sections:
            if section!='':
                sections_newlist.append(section)

        # add all the properties
        title           = request.title
        fromname	= request.fromname
        email    	= request.email
        url2issue 	= request.url2issue
        type     	= request.type
        urgency 	= request.urgency
        description 	= request.description
        display_format  = request.display_format
        confidential 	= int(request.confidential)
        hide_me 	= int(request.hide_me)
        issuedate 	= getNowdate()
        status  	= request.get('status','open')
        sections        = sections_newlist


        # Do the actual object adding
        issue = self._createIssueObject(genid, request.title, status, type, urgency,
                                  sections, fromname, email, url2issue,
                                  confidential, hide_me, description,
                                  display_format, issuedate)
        

        # catalog it
        issue.index_object()

        # Whereto next?
        redirect_url = '%s?NewIssue=Submitted'%(issue.absolute_url())
        request.RESPONSE.redirect(redirect_url)


    def _createIssueObject(self, id, title, status, type, urgency, sections,
                     fromname, email, url2issue, confidential, hide_me,
                     description, display_format, issuedate, index=0):
        """ crudely create issue object. No checking """
        issue = IssueTrackerIssue(id, title, status, type, urgency,
                                  sections, fromname, email, url2issue,
                                  confidential, hide_me, description,
                                  display_format, issuedate)

        # not here
        self._setObject(id, issue)
        issue = getattr(self, id)

        if index:
            # catalog it
            issue.index_object()

        return issue

    def _generateID(self, start, length, prefix='',
                    meta_type='Issue Tracker Issue',increment=None):
        """ generate IDs for different objects """
        if increment is None:
            idnr = len(self.objectIds(meta_type))+1
            increment = idnr + 1
        else:
            increment = increment +1
            idnr = increment
        id=string.zfill(str(idnr), length)
        if hasattr(self, id):
            # ah! Id already exists, try again
            return self._generateID(start, length, \
                                    meta_type, increment=increment)
        else:
            return id
        

    def About(self):
        """ Show some info about the product """
        osj = os.path.join
        About_content = open(osj(package_home(globals()), 'About.html')).read()
        About_content = About_content.replace('*CHANGELOG*',\
                            ShowDescription(CHANGELOG, 'structured_text'))
        About_content = About_content.replace('*VERSION*',__version__)
        return About_content

    # Cookies!

    def get_cookie(self, name):
        """ return RESPONSE cookie """
        return self.REQUEST.cookies[name]
    
    def set_cookie(self, name, value):
        """ set RESPONSE cookie """
        later = str(getLaterDate(365))
        self.REQUEST.RESPONSE.setCookie(name,value, expires=later, path='/')
        
    def has_cookie(self, name):
        """ return cookie presence """
        return self.REQUEST.cookies.has_key(name)


    # Misc

    def preParseEmailString(self, email_string):
        """ wrapper around utils """

        parsemethod = IssueTrackerUtils.preParseEmailString
        if self.hasNotifyables():
            names2emails = {}
            for item in self.getNotifyables():
                names2emails[item.getName()] = item.getEmail()
            return parsemethod(email_string, names2emails=names2emails)
        else:
            return parsemethod(email_string)
        
    def hasManagerRole(self):
        """
            This method determines if the current user
            is allowed to do stuff that only the Zope manager is
            supposed to be able to do.
            Feel free to edit appropriatly to what suits you.
        """
        return 'Manager' in self.REQUEST.AUTHENTICATED_USER.getRoles()

    def getRandomString(self, length=5, loweronly=0):
        """ return a completely random piece of string """
        return IssueTrackerUtils.getRandomString(length, loweronly)

    
    def thisInURL(self, page, homepage=0):
        """ To find if a certain objectid is in the URL """
        URL = self.REQUEST.URL
        if page in URL.split('/'):
            if homepage:
                if string.find(URL.split('/')[-2],'Issue') > -1:
                    return 0
            return '#CCCCCC'
        else:
            return 0

    def ActionURL(self, url=None):
        """
        If URL is http://host/index_html
        I prefer to display it http://host
        Just a little Look&Feel thing
        """
        if url is None:
            url = self.REQUEST.URL

        URLsplitted = string.split(url,'/')
        if URLsplitted[-1] == 'index_html':
            return string.join(URLsplitted[:-1],'/')

        return url

    def ExplainSubmitErrors(self, dict):
        """ Return HTMLed statements about errors """
        li = []
        for key,value in dict.items():
            li.append('<strong>%s</strong> - %s'%(key, value))
        return li

    def ShowFilter(self, filtername, sequence=[]):
        """ NOT SURE OF THIS METHODS USE"""
        request = self.REQUEST
        if request.has_key(filtername):
            return request[filtername]
        elif not request.has_key('filteroptions'):
            return sequence

        return []

    def ShowOnlyMe(self):
        """ enable listing with set name and email """
        request = self.REQUEST
        redirect = request.RESPONSE.redirect

        filters={}
        for each in ['fromname','email']:
            if request.has_key(each) and request[each] != '':
                filters[each]=request[each]

        if filters:
            querystring = []
            for k,v in filters.items():
                querystring.append('%s=%s'%(k,v))
            # make list into string
            querystring = '?'+string.join(querystring,'&')
        else:
            querystring=''

        redirect(self.ListIssues.absolute_url()+querystring,lock=1)

    def ShowNameEmail(self, fromname, email, hideme=None):
        """ Show name and email depending on certain criterias """
        out = ''
        fromname = html_quote(fromname)
        email    = html_quote(email)

        if fromname =='' and email=='':
            name_email = NONAME_NOEMAIL
        elif fromname =='': 
            name_email ='&lt;<a href="mailto:%s">%s</a>&gt;'%(email,email)
        elif email =='':
            name_email = fromname
        else:
            name_email = '%s &lt;<a href="mailto:%s">%s</a>&gt;'%(fromname, email, email)
         

        if hideme is not None and hideme:
            out += NAME_EMAIL_HIDDEN    
            if self.hasManagerRole():
                out += name_email
        else:
            out += name_email

        return out

    def has_key_special(self, name):
        """
        Normally you would do REQUEST.has_key('IssueAction')
        but if an imagebutton is used you'll find that you have
        REQUEST['IssueAction.y'] and REQUEST['IssueAction.x']
        But the result should be the same.
        """
        request = self.REQUEST
        if request.has_key(name):
            return 1
        elif request.has_key('%s.y'%name) and request.has_key('%s.x'%name):
            return 1
        else:
            return 0

    def get_special_key(self, name):
        """
        Normally you would do REQUEST.has_key('IssueAction')
        but if an imagebutton is used you'll find that you have
        REQUEST['IssueAction.y'] and REQUEST['IssueAction.x']
        But the result should be the same.
        """
        try:
            return self.REQUEST[name]
        except KeyError:
            try:
                return self.REQUEST['%s.x'%name]
            except:
                raise KeyError, name

    def getSavedUser(self, name_email='email'):
        """
        Return the name or email from request,
        if not found, return from the session object,
        else return ""
        """
        request = self.REQUEST

        if name_email =='email':
            s='email'
            cookie=self.email_cookiekey
        else:
            s='fromname'
            cookie=self.name_cookiekey

        # now we know what we're looking for

        if request.has_key(s):
            return request[s]
        elif self.has_cookie(cookie):
            return self.get_cookie(cookie)

        return ""


    def ManagerOptionsExtend(self):
        """ Determine which forms to show to Managers """
        request = self.REQUEST

        has_key_special = self.has_key_special
        get_special_key = self.get_special_key

        if has_key_special('IssueAction'):
            action = get_special_key('IssueAction')
            print "I'm here and action is: ", action
            print getattr(self,request['issueID']).absolute_url()
            return getattr(self,request['issueID']).ModifyIssue(action=action)
        elif has_key_special('IssueAction_AddFollowup'):
            return self.form_followup()
        elif has_key_special('IssueAction_Open'):
            return self.form_open()
        elif has_key_special('IssueAction_Take'):
            return self.form_take()
        elif has_key_special('IssueAction_Onhold'):
            return self.form_onhold()
        elif has_key_special('IssueAction_Rejected'):
            return self.form_reject()
        elif has_key_special('IssueAction_Complete'):
            return self.form_complete()
        elif has_key_special('IssueAction_Delete'):
            return self.form_delete()
        else:
            return self.manager_options()
        return ""

    def AnonymousOptionsExtend(self):
        """ Determine which forms to show to Anonymous """
        request = self.REQUEST

        if self.has_key_special('IssueAction'):
            action=self.get_special_key('IssueAction')
            return getattr(self,request['issueID']).ModifyIssue(action=action)
        elif self.has_key_special('IssueAction_AddFollowup'):
            return self.form_followup()
        else:
            return self.anonymous_options()

        return ""


    def CountStatuses(self, since=None):
        """ Return how many Issues there are under each status since
            a certain time
        """
        return {}
        request = self.REQUEST

        # check that since isn't a string
        if same_type(since,'s'):
            since = DateTime(since)
        elif request.has_key('count_status_since'):
            since = DateTime()-int(req['count_status_since'])

        res={}

        for issue in self.objectValues('Issue Tracker Issue'):
            if since is None or issue.bobobase_modification_time() > since:
                if res.has_key(issue.status):
                    res[issue.status]=res[issue.status]+1
                else:
                    res[issue.status] = 1

        return res

    def BatchedQueryString(self, batchdict):
        """
        return QUERY_STRING but make sure stuff in
        the batchdict isn't duplicated.
        """
        request = self.REQUEST
        qs = request.QUERY_STRING

        old_qs ={}
        if qs <> '':
            for pair in qs.split('&'):
                pair_split = pair.split('=',1)
                if pair_split[0] not in batchdict.keys():
                    old_qs[pair_split[0]] = pair_split[1]

        new_qs = ""
        for k,v in batchdict.items():
            new_qs = new_qs + '&%s=%s'%(k,v)
        for k,v in old_qs.items():
            new_qs = new_qs + '&%s=%s'%(k,v)

        return new_qs.replace('&','?',1)

    def DeleteIssue(self):
        """ Delete an Issue from the IssueTracker instance """
        request = self.REQUEST

        if request.has_key('issueID') and self.hasManagerRole():
            self.manage_delObjects(request['issueID'])

            # delete all notifications about this Issue
            del_notify_ids = []
            for notifyobject in self.objectValues('Issue Notification'):
                if notifyobject.issueID == request['issueID']:
                    del_notify_ids.append(notifyobject.id)
          
            self.manage_delObjects(del_notify_ids)

            request.RESPONSE.redirect(request.URL1+'/ListIssues')
        else:
            raise "IssueTrackerError", "The issueID could not be found in the REQUEST"

    def standard_html_header(context):
        """ to make it possible to use DTML objects here """
        breakword = '<!--METALbody-->'
        page = context.StandardLook()
        return page[:page.find(breakword)]

    def standard_html_footer(context):
        """ to make it possible to use DTML objects here """
        breakword = '<!--METALbody-->'
        page = context.StandardLook()
        return page[page.find(breakword)+len(breakword)+1:]

    def ManagerLink(context):
        """ For the little hyperlink where you can login with """
        link = context.login.absolute_url()
        came_from = context.ActionURL()
        rurl=str(int(getNowdate()))[0:4]
        return "%s?came_from=%s&r.url=%s"%(link, came_from, rurl)

    def getSavedTextFormat(self, no_default=0):
        """
        This method returns what display_format value the user has. 
        If none found, then the default one is returned.
        """

        if no_default:
            default = ""
        else:
            default = self.default_display_format

        s=None
        cookie = self.display_format_cookiekey

        if self.has_cookie(cookie):
            s = self.get_cookie(cookie)

        if s not in self.display_formats:
            s = None

        if s is None:
            return default
        else:
            return s


    # Notifyables
    #


    def hasNotifyables(self):
        """ see if there are any notifyables at all """
        if len(self.notifyables.objectValues('Issue Tracker Notifyable')) > 0:
            return 1
        else:
            return 0
        
    def getNotifyables(self):
        """ Return all Notifyable objects """
        return self.notifyables.objectValues('Issue Tracker Notifyable')
    
    def manage_editNotifyables(self, REQUEST):
        """ Save notifyables """
        updates = []

        for notify_id in REQUEST.get('notify_ids',[]):
            alias = REQUEST['alias_%s'%notify_id].strip()
            email = REQUEST['email_%s'%notify_id].strip()
            if ValidEmailAddress(email):
                updates.append({'id':notify_id,
                                'email':email,
                                'alias':alias})                                
        for c in range(len(REQUEST.get('new_email',[]))):
            email = REQUEST['new_email'][c]
            alias = REQUEST['new_alias'][c]
            try:
                groups = self._filterNotifyGroups(REQUEST['new_groups'])
            except: groups=[]
            if type(groups) != type([]):
                groups = [groups]
            if ValidEmailAddress(email):
                updates.append({'id':GenerateNotifyableId(email),
                                'email':email,
                                'alias':alias,
                                'groups':groups})
                
        self._saveNotifyables(updates)
        return self.manage_ManagementForm(REQUEST,
                        manage_tabs_message='Notifyables updated.')


    def _saveNotifyables(self, updates):
        """ save the existing notifyables """
        for d in updates:
            id, alias, email = d['id'], d['alias'], d['email']
            if d.has_key('groups'):
                groups = d['groups']
            else:
                groups = None
            if hasattr(self.notifyables,id):
                # update
                n = self.notifyables[id]
                n.alias = alias.strip() 
                n.email = email.strip()
                if groups is not None:
                    n.groups = self._filterNotifyGroups(groups)
                else:
                    n.groups = self._filterNotifyGroups(n.groups)
            else:
                # create
                n = IssueTrackerNotifyable(id, alias, email,
                                           self._filterNotifyGroups(groups))
                self.notifyables._setObject(id, n)
                
                

    def manage_delNotifyables(self, REQUEST):
        """ Prepare which notifyable ids to remove """
        ids = REQUEST.get('del_notify_ids',[])
        self.notifyables.manage_delObjects(ids)
            
        return self.manage_ManagementForm(REQUEST,
                        manage_tabs_message='Notifyables deleted.')


    def manage_editNotifyable(self, id, alias=None, email=None, groups=None,
                              REQUEST=None):
        """ save changes to Notifyable """
        no = self.notifyables[id]
        no.manage_editNotifyable(alias, email, groups, REQUEST)

    def _filterNotifyGroups(self, groups):
        """ Return the list groups but filter out groups that
            don't exists in self.notify_groups
        """
        n_groups=[]
        for group in groups:
            if group.strip() in self.notify_groups and \
               group.strip() not in n_groups:
                n_groups.append(group.strip())
        return n_groups
        

    # Notify groups
    #

    def manage_delNotifyGroups(self, notify_groups, REQUEST=None):
        """ delete some groups from self.notify_groups """
        new_groups = []
        for old in self.notify_groups:
            if old not in notify_groups:
                new_groups.append(old)
        self.notify_groups = new_groups

        # fix things up
        self._removeGroupsfromUsers()
        
        return self.manage_ManagementForm(REQUEST,
                        manage_tabs_message='Notifygroups deleted.')

    def manage_saveNotifyGroup(self, notify_group, REQUEST=None):
        """ add one group """
        self.notify_groups.append(notify_group)
        return self.manage_ManagementForm(REQUEST,
                        manage_tabs_message='Notifygroup added.')

    def _removeGroupsfromUsers(self):
        """ go through all users and make sure that they only belong
            to groups that exist """
        ok_groups = self.notify_groups
        notifyables = self.getNotifyables()
        for notifyable in notifyables:
            notifyable.groups = self._filterNotifyGroups(notifyable.groups)

    # List issues
    #

    def ListIssuesFiltered(self):
        """ wrapper around _ListIssuesFiltered() that prepares a search
        if REQUEST holds 'q'
        """
        q=None
        catalog = self.Catalog
        if self.REQUEST.has_key('q') and self.REQUEST['q'] != '':
            q=self.REQUEST['q'].strip()
##            q=q.replace('%','*') # allow both wildcards
        seq = []
        for brain in catalog():
            seq.append(brain.getObject())
        return self._ListIssuesFiltered(seq)
        
##            catalogs = catalog({'title':q}) + \
##                       catalog({'description':q}) + \
##                       catalog({'fromname':q}) + \
##                       catalog({'email':q})
##            seq=[]
##            for brain in catalogs:
##                try:
##                    issueobject = getattr(self,brain.id)
##                    if issueobject not in seq:
##                        seq.append(issueobject)
##                except:
##                    # probably because issue has been deleted manually
##                    pass
##        else:
##            seq = self.objectValues('Issue Tracker Issue')
##        return self._ListIssuesFiltered(seq)

    def _ListIssuesFiltered(self,objsequence):
        """ Basically filter a sequence of issue objects """
        rlist = []
        request = self.REQUEST

        set_cookie = self.set_cookie

        try:
            # if there and integer'able
            reverse=int(request.reverse)
        except:
            reverse=0

        trackerroot = self.getRoot()

        #TrackerFilteroptionsCookiename
        tfc = "%s_filteroptions"%trackerroot.id
        
        # remember filteroptions
        if request.has_key('filteroptions'):
            tracker_filteroptions={}
            for each in ['ignoretypes','ignoresections','ignorestatuses',\
                         'ignorefromname','ignoreurgencies','ignoreemail']:
                try:
                    tracker_filteroptions[each]=request[each]
                except:pass
            set_cookie(tfc,tracker_filteroptions)
            
       
        #
        # prepare a master filter
        #
        
        filter ={}
        
        # statuses
        if request.has_key('ignorestatuses'):
            filter['status'] = request.ignorestatuses
        elif request.has_key('status'):
            filter['status'] = [request.status.strip()]
        else:
            filter['status'] = []
            
        # sections
        if request.has_key('ignoresections'):
            filter['sections'] = request.ignoresections
        elif request.has_key('section'):
            filter['sections'] = [request.section.strip()]
        else:
            filter['sections'] = []

        # urgencies
        if request.has_key('ignoreurgencies'):
            filter['urgency'] = request.ignoreurgencies
        elif request.has_key('urgency'):
            filter['urgency'] = [request.urgency.strip()]
        else:
            filter['urgency'] = []
            
        # types
        if request.has_key('ignoretypes'):
            filter['type'] = request.ignoretypes
        elif request.has_key('type'):
            filter['type'] = [request.type.strip()]
        else:
            filter['type'] = []
            
        ## fromname
        # Either a name to ignore or show only.
        # It does not make sense to have both showfromname and ignorefromname
        if request.has_key('showfromname') and \
           request.showfromname <> self.when_ignore_word:
                filter['showfromname'] = request.showfromname.strip()
        elif request.has_key('ignorefromname') and \
             request.ignorefromname <> self.when_ignore_word:
            filter['ignorefromname'] = request.fromname.strip()

        # email
        if request.has_key('showemail') and \
           request.showemail <> self.when_ignore_word:
                filter['showemail'] = string.strip(request.showemail)
        elif request.has_key('ignoreemail') and \
           request.ignoreemail <> self.when_ignore_word:
            filter['ignoreemail'] = request.email


        # prepare for caseindependence
        filter = fixDictofLists(filter)
            
        # all Issue objects
        issues = objsequence 

        #
        # Before we can determine what to sort on
        # we need to check the name of the properties
        
        # the default
        sortorder = ('bobobase_modification_time',)
        skipsort = 0
        if request.has_key('sortorder') and request['sortorder'] != 'date':
            sortorder = (request['sortorder'],)
            if request['sortorder'] == 'from':
                sortorder = ('fromname',)
            if request['sortorder'] == 'creationdate':
                sortorder = ('issuedate',)
            if request['sortorder'] == 'changedate':
                sortorder = ('bobobase_modification_time',)

            if request['sortorder'] == 'status':
                #
                # Very special!
                # self.statuses is an "human ordered" list like 'open','taken',...'completed','forgotten'
                statuses = {}
                # create a dictionary of list divided in groups by their status
                # like this: {'open':['foo','bar'], ... 'forgotten':['this','that']}
                for issue in issues:
                    if statuses.has_key(issue.status):
                        statuses[issue.status].append(issue)
                    else:
                        statuses[issue.status] = [issue]
                # recreate the containers
                issues = []
                # remembers that self.statuses is ordered
                #
                # self.statuses is the property
                # statuses is the dictionary
                #
                for status in self.statuses:
                    # perhaps there are no 'forgotten' issues even
                    # though there is 'forgotten' in self.statuses
                    if statuses.has_key(status):
                        # order all issues with this status by bobobase_modification_time
                        thisstatus_containers = sequence.sort(statuses[status],(('bobobase_modification_time',),))
                        # reverse bobobase_modification_time sort naturally
                        thisstatus_containers.reverse()
                        issues = issues + thisstatus_containers
                        # skip the default sort later
                        skipsort = 1
                        
            if request['sortorder'] == 'urgency':
                #
                # Very special too
                # read the comments on "if request['sortorder'] == 'status':"
                # above.
                urgencies = {}
                for issue in issues:
                    if urgencies.has_key(issue.urgency):
                        urgencies[issue.urgency].append(issue)
                    else:
                        urgencies[issue.urgency] = [issue]
                issues = []
                for urgency in self.urgencies+['']:
                    if urgencies.has_key(urgency):
                        thisurgency_containers = sequence.sort(urgencies[urgency], (('bobobase_modification_time',),))
                        thisurgency_containers.reverse()
                        issues = issues + thisurgency_containers
                        skipsort = 1
                
        else:
            # if no sortorder, it uses the bobobase_modification_time
            # and this has to be reversed so that the latest is listed first
            reverse =1

        # do the basic sort.
        # 'sortorder' is a tuple
        if not skipsort:
            issues = sequence.sort(issues,(sortorder,))

            
        if reverse:
            issues.reverse()

        # containers is not order, just filter it
        for issue in issues:
            # if confidential and (if) Manager show, if not skip
            if not issue.IsConfidential() or self.hasManagerRole():
                if filter == {}:
                    rlist.append(issue)
                else:
                    passsections = 1
                    for section in issue.sections:
                        if ss(section) in filter['sections']:
                            passsections = 0
                            break
                    # pass the filter lists

                    if ss(issue.status) not in filter['status'] \
                       and ss(issue.urgency) not in filter['urgency'] \
                       and ss(issue.type) not in filter['type'] \
                       and passsections and \
                       self._filter_fromnameemail(issue, filter):
                        rlist.append(issue)

                        # last property check.
                        # If filter has ignorefromname, 
                        
##        
##                        if (ss(con.fromname) == filter['fromname'] or filter['fromname']=='') \
##                           and \
##                           (ss(con.email) == filter['email'] or filter['email'] =='' ):
##                            rlist.append(con)
                
        return rlist

    def _filter_fromnameemail(self, issue, filter):
        """ do the advanced filtering.
            If filter has ignorefromname then reject this issue.
            If filter has showfromname, then ignore this issue
            if not == showfromname
        """
        # if asked to be ingored return false
        if filter.has_key('ignorefromname') and filter.ignorename !='' and \
           ss(issue.fromname) == filter['ignorefromname']:
            return 0
        elif filter.has_key('ignoreemail') and filter.ignoreemail !='' and \
           ss(issue.email) == filter['ignoreemail']:
            return 0
        elif filter.has_key('showfromname') and filter.showfromname !='' and \
           ss(issue.fromname) != filter['showfromname']:
            return 0
        elif filter.has_key('showemail') and filter.showemail !='' and \
           ss(issue.email) != filter['showemail']:
            return 0
        else:
            # not trapped in anything
            return 1

    def HeadingLinks(self, display, sortname, default=0, startreversed=0):
        """Returns a hyperlink that can
           be used for resorting the listing
        """
        request = self.REQUEST
        querystring = request.QUERY_STRING

        ascending = self.www['ascarrow.gif'].tag(hspace=2)
        descending = self.www['descarrow.gif'].tag(hspace=2)
        invisible = self.www['invisiblearrow.gif'].tag(hspace=2)
        
        
        n_querystring = []
        for query in string.split(querystring,'&'):
            if string.find(query,'sortorder') == -1 and string.find(query,'reverse') == -1:
                n_querystring.append(query)
        querystring = string.join(n_querystring,'&')

        if querystring != '':
            querystring = '&%s'%querystring
        
        if request.has_key('reverse') and (request['reverse'] == 1 or request['reverse'] == '1'):
            reverse = 1
        elif startreversed and not (request.has_key('sortorder') and request['sortorder']==sortname):
            reverse = 1
        else:
            reverse = 0
        if request.has_key('sortorder') and request['sortorder'] == sortname or (default and not request.has_key('sortorder')):
            # have sorted by this, just let them reverse
            if reverse:
                return '%s<a class="headinglink_current" href="%s?sortorder=%s%s" title="%s %s">%s</a>' \
                       % (descending,request.URL, sortname, querystring, SORT_BY, display, display )
            else:
                return '%s<a class="headinglink_current" href="%s?sortorder=%s&reverse=1%s" title="%s">%s</a>'\
                       % (ascending,request.URL, sortname, querystring, SORT_REVERSE, display )
        else:
            if startreversed:
                return '%s<a class="headinglink" href="%s?sortorder=%s%s&reverse=1" title="%s %s">%s</a>' \
                   % (invisible,request.URL, sortname, querystring, SORT_BY, display, display )
            else:
                return '%s<a class="headinglink" href="%s?sortorder=%s%s" title="%s %s">%s</a>' \
                   % (invisible,request.URL, sortname, querystring, SORT_BY, display, display )
        
    # Templates
    #
    
    PTF = PageTemplateFile

##    StandardLook = PTF('zpt/StandardLook', globals(), __name__='StandardLook')
    AddIssue = PTF('zpt/AddIssue', globals())
    QuickAddIssue = PTF('zpt/QuickAddIssue', globals())
    SearchForm = PTF('zpt/SearchForm', globals())
    user_searchform = PTF('zpt/user_searchform', globals())
    submitform = PTF('zpt/submitform', globals())
    quickpreviewform = PTF('zpt/quickpreviewform', globals())
    previewform = PTF('zpt/previewform', globals())
    latestIssues = PTF('zpt/latestIssues', globals())
    ListIssues = PTF('zpt/ListIssues', globals(), __name__='ListIssues')
#    listissues._setName('listissues')
    index_html = PTF('zpt/index_html', globals())
    ShowIssue = PTF('zpt/ShowIssue', globals())
    ShowIssueData = PTF('zpt/ShowIssueData', globals())
    
    # Start StandardLook once
##    StandardLook()
    
 
IssueTracker.__dict__['About.html'] = IssueTracker.About
#IssueTracker.__dict__['listissues'] = IssueTracker.ListIssues
InitializeClass(IssueTracker)
    

###############   ISSUETRACKERISSUE    ###################

class IssueTrackerIssue(IssueTracker):
    """ Issues class as containers """

    meta_type = 'Issue Tracker Issue'
    icon = '%s/issue.gif'%ICON_LOCATION
    
    _properties=({'id':'title',         'type': 'string', 'mode':'w'},
                 {'id':'issuedate',     'type': 'date',   'mode':'w'},
                 {'id':'status',        'type': 'string', 'mode':'w'},
                 {'id':'type',          'type': 'string', 'mode':'w'},
                 {'id':'urgency',       'type': 'string', 'mode':'w'},
                 {'id':'sections',      'type': 'lines',  'mode':'w'},
                 {'id':'fromname',      'type': 'string', 'mode':'w'},
                 {'id':'email',         'type': 'string', 'mode':'w'},
                 {'id':'url2issue',     'type': 'string', 'mode':'w'},
                 {'id':'confidential',  'type': 'boolean','mode':'w'},
                 {'id':'hide_me',       'type': 'boolean','mode':'w'},
                 {'id':'description',   'type': 'text',   'mode':'w'},
                 {'id':'display_format','type': 'string', 'mode':'w'},
                 )
    
    security=ClassSecurityInfo()

    manage_options = (
        {'label':'Contents', 'action':'manage_main'},
        {'label':'Properties', 'action':'manage_propertiesForm'}
        )


    def __init__(self, id, title, status, issuetype, urgency, sections,
                 fromname, email, url2issue, confidential, hide_me,
                 description, display_format, issuedate=''):
        """ init an Issue object """
        self.id = str(id)
        self.title = title.strip()
        if type(issuedate) == type('s'):
            issuedate = DateStr2DateDate(issuedate)
        self.issuedate = issuedate
        self.status = status
        self.type = issuetype
        self.urgency = urgency
        self.sections = sections
        self.fromname = fromname
        self.email = email
        self.url2issue = url2issue
        self.confidential = confidential
        self.hide_me = hide_me
        self.description = description
        if display_format:
            self.display_format = display_format
        else:
            self.display_format = self.default_display_format
        
    security.declareProtected('View', 'index_html')
    def index_html(self, REQUEST=None):
        """ show the issue """
        if not self.confidential or self.hasManagerRole():
            return self.ShowIssue()
        else:
            self.REQUEST.RESPONSE.redirect(self.ListIssues.absolute_url())

    def IsConfidential(self):
        """ check confidential property """
        if hasattr(self, 'confidential') and self.confidential:
            return 1
        else:
            return 0
        
    def ModifyIssue(self,action='Addfollowup'):
        """ advanced change to issue properties """
        request = self.REQUEST
        
        action = action.strip().lower()
        oldstatus = self.status
        prefix = self.issueprefix
                
        if not request.has_key('comment_description'):
            request.set('comment_description','')

        if not request.has_key('display_format'):
            request.set('display_format',self.default_display_format)

        req_manager = 1 # require manager by default
        dummychange = 0
        addfollowup = 0
        past_tense=None
        if action == 'open':
            past_tense='open'
            addfollowup = 1
        if action == 'take':
            past_tense='taken'
            addfollowup = 1
        if action == 'put on hold':
            past_tense='on hold'
            addfollowup = 1
        if action == 'reject':
            past_tense='rejected'
            addfollowup = 1
        if action == 'complete':
            past_tense='completed'
            addfollowup = 1
        if action == 'delete':
            DeleteIssue(self, self.id)
            addfollowup = 0
            redirect_url = request.URL2
        if action == 'addfollowup':
            addfollowup = 1
            dummychange = 1
            req_manager = 0

        if req_manager and not self.hasManagerRole():
            # the chosen action requires manager role,
            # but the person is not a manager
            self.login(self,request)

        if dummychange:
            # Still should trigger a new bobobase_modification_time
            status = self.status
            
        if past_tense is not None:
            self.status = past_tense
        else:
            # anything else means that the comment_description
            # cannot be empty
            if request['comment_description'].strip()=='':
                addfollowup=0
                redirect_url = request.URL1
        
        # most actions may perhaps add a little comment
        if addfollowup:

            genid = self._generateID("IssueThread", self.randomid_length, \
                               prefix, meta_type='Issue Tracker Issue Thread')
            
            if action == 'addfollowup':
                gentitle = 'Added Issue followup'
            else:
                gentitle = 'Changed status from %s to %s'%\
                           (oldstatus.capitalize(), past_tense.capitalize())

            # fix variables
            try:
                fromname = request.fromname
            except:
                fromname = ''
            try:
                email = request.email
            except:
                email = ''
            try:
                display_format = request.display_format
            except:
                display_format = self.default_display_format

                
            # remember visitor
            self.set_cookie(self.name_cookiekey,fromname)
            self.set_cookie(self.email_cookiekey,email)
            self.set_cookie(self.display_format_cookiekey,display_format)

            # update Thread object
            title  = gentitle
            comment = request.comment_description
            threaddate = getNowdate()

            # create an Thread object
            followupobject = self._createThreadObject(genid, title, comment,
                                                      threaddate, fromname,
                                                      email, display_format)
            
            if request.has_key('notify'):
                        
                # now, create and email-alert-queue object
                # using filtered email address
                # get who to notify        
                email_addresses = self.Others2Notify(genid, do='email',emailtoskip=email)

                if len(email_addresses) > 0:

                    # get root instance
                    notifyid = self._generateID('Notification',5, prefix,
                                                meta_type='Issue Notification')
                    
                    title = self.title
                    change = gentitle
                    issueID = self.id
                    anchorname = len(self.objectIds('Issue Tracker Issue Thread'))
                    emails = email_addresses
                    date = getNowdate()
                    if not hasattr(self, issueID):
                        raise "IssueNonexisting", "This notification has no issue"
                    else:
                        notification = IssueTrackerNotification(
                                        notifyid, title, change, issueID,
                                        comment, emails, anchorname, fromname,
                                        date)
                        self._setObject(notifyid, notification)
                        notifyobject = getattr(self, notifyid)
                      
                        # use the dispatcher to try to send
                        # this notification right now.
                        request.set('loop_over',[notifyobject])
                        # there is no big deal if the dispatcher crashes here
                        # because the notification is saved and the dispatcher
                        # can be invoked some other time manually
                        try:
                            getattr(self,'dispatcher.html')(self,req)
                        except:
                            LOG(self.meta_type, PROBLEM, 'Email could not be sent', 'error=sys.exc_info()')
                            #pass
            
            
            redirect_url = '%s#i%s'%\
                           (self.absolute_url(),len(self.objectIds('Issue Tracker Issue Thread')))
        # catalog
        self.reindex_object()
        # ready ! redirect!
        request.RESPONSE.redirect(redirect_url)

    def _createThreadObject(self, id, title, comment, threaddate,
                            fromname, email, display_format):
        """ Crudely create thread object. No checking. """
        thread = IssueTrackerIssueThread(id, title, comment, threaddate,
                                         fromname, email, display_format)
        self._setObject(id, thread)

        # get that object
        threadobject = getattr(self, id)

        return threadobject
    
    def Others2Notify(self, issueID, do='print',emailtoskip=None):
        """
        Returns a list of names and emails of guys to notify on this Issue. 
        Take the hide_me and confidential factor into consideration.
        
        if do == 'print':
            display the people on screen (using hide_me and confidential factor)
        if do == 'email':
            return a pure list of all email address
        """

        if emailtoskip is None:
            if self.has_cookie(self.email_cookiekey):
                emailtoskip = self.get_cookie(self.email_cookiekey)
            else:
                emailtoskip = '@@@@@'

        fromstrings = []


        issue = self

        # proceed only if Manager or open-name on the issue
        if self.hasManagerRole() or not issue.hide_me:
            # self.email is the email of the thread
            # emailtoskip comes from the param/cookie if applicable
            if not issue.email == emailtoskip:
                if do =='email':
                    fromstring = getFromstring(issue.fromname, issue.email, email_must=1, email_pure=1)
                else:
                    fromstring = getFromstring(issue.fromname, issue.email)
                # fromstring looks like this now: "Peter Bengtsson <mail@peterbe.com>"
                # or ['mail@peterbe.com'] depending on 'do'
                # ...maybe
                if fromstring and fromstring not in fromstrings:
                    fromstrings.append(fromstring)

        for thread in issue.objectValues('Issue Tracker Issue Thread'):
            if not thread.email == emailtoskip:
                if do =='email':
                    fromstring = getFromstring(thread.fromname, thread.email, email_must=1, email_pure=1)
                else:
                    fromstring = getFromstring(thread.fromname, thread.email)
                if fromstring and fromstring not in fromstrings:
                        fromstrings.append(fromstring)
        return fromstrings
    

    security.declareProtected('View', 'issue2TXT')
    def issue2TXT(self, RESPONSE):
        """ prepare an issue for download """
        if not self.confidential or self.hasManagerRole():

            if RESPONSE is not None:
                RESPONSE.setHeader('Content-type','application/data')
                RESPONSE.setHeader('Content-Disposition','inline;filename=%s.txt' % (self.id))

            return self.issue2TXT_script(self)
        else:
            self.REQUEST.RESPONSE.redirect(self.ListIssues.absolute_url())

    security.declareProtected('Access contents information', 'issue2TXT_script')
    def issue2TXT_script(self,issue):
        """ the textual representation of an issue """

        nl="\r\n"
        TXT =""
        TXT += "Export of issue (%s) %s"%(issue.id,
                                          self.ZopeTime().strftime(self.display_date))
        TXT += "%s--------------------------------%s%s"%(nl,nl,nl)
        TXT += " %s%s%s"%(issue.title,nl,nl)
        TXT += "Status:     %s%s"%(issue.status.capitalize(),nl)
        TXT += "Issue date: %s%s"%(issue.issuedate.strftime(self.display_date),nl)
        TXT += "Type:       %s%s"%(issue.type,nl)
        TXT += "Urgency:    %s%s"%(issue.urgency,nl)
        TXT += "Sections:   %s%s"%(string.join(issue.sections,', '),nl)
        TXT += "URL:        %s%s"%(issue.url2issue,nl*2)

        if issue.hide_me:
            TXT += "Name and email hidden%s"%nl
            if context.hasManagerRole():
                TXT += "Fromname:   %s%s"%(issue.fromname,nl)
                TXT += "Email:      %s%s"%(issue.email,nl*2)
        else:
            TXT += "Fromname:   %s%s"%(issue.fromname,nl)
            TXT += "Email:      %s%s"%(issue.email,nl*2)

        TXT += nl*2
        TXT += "%s%s"%(issue.description.replace('\n',nl),nl)

        threads = issue.objectValues('Issue Tracker Issue Thread')
        if len(threads)>0:
            TXT += "%s------------Followups------------%s"%(nl*2,nl*2)
            for thread in threads:
                TXT += "%s %s%s"%(nl,thread.title,nl*2)
                
                TXT += "Followup date: %s%s"%(thread.threaddate.strftime(self.display_date),nl)
                TXT += "Fromname:      %s%s"%(thread.fromname,nl)
                TXT += "Email:         %s%s"%(thread.email,nl*2)

                TXT += thread.comment.replace('\n',nl)
                TXT += nl*2

        return TXT

    def getOptionButtons(self):
        res=[]

        img_con = self.actionbuttons

        if self.status !='open':
            res.append(['IssueAction_Open','Open', \
                        img_con['open.gif'].absolute_url()])
        if self.status !='taken':
            res.append(['IssueAction_Take','Take', \
                        img_con['take.gif'].absolute_url()])
        if self.status != 'on hold':
            res.append(['IssueAction_Onhold','Put on hold', \
                        img_con['putonhold.gif'].absolute_url()])    
        if self.status != 'rejected':
            res.append(['IssueAction_Rejected','Reject', \
                        img_con['reject.gif'].absolute_url()])    
        if self.status != 'completed':
            res.append(['IssueAction_Complete','Complete', \
                        img_con['complete.gif'].absolute_url()])
        res.append(['IssueAction_Delete','Delete', \
                    img_con['delete.gif'].absolute_url()])

        return res

    def ListThreads(self,issueID):
        """ Return the followup objects. The threads. """
        req = self.REQUEST
        rlist = []
        issueobject = getattr(self,issueID)

        threads = issueobject.objectValues('Issue Tracker Issue Thread')
        for thread in threads:
            # if testing on the thread needs to be done,
            # do that here.
            rlist.append(thread)
            
        return rlist
    
    security.declareProtected('Manage ZCatalog Entries', 'manage_beforeDelete')
    def manage_beforeDelete(self, REQUEST=None, RESPONSE=None):
        """ uncatalog yourself """
        self.unindex_object()

    # Templates
    PTF = PageTemplateFile
    
    shownewIssueURL = PTF('zpt/shownewIssueURL', globals())
    manager_options = PTF('zpt/manager_options', globals())
    anonymous_options = PTF('zpt/anonymous_options', globals())
    form_followup = PTF('zpt/form_followup', globals())
    form_open = PTF('zpt/form_open', globals())
    form_take = PTF('zpt/form_take', globals())
    form_onhold = PTF('zpt/form_onhold', globals())
    form_reject = PTF('zpt/form_reject', globals())
    form_delete = PTF('zpt/form_delete', globals())
    form_complete = PTF('zpt/form_complete', globals())


InitializeClass(IssueTrackerIssue)


############### ISSUETRACKERISSUETHREAD ##################

class IssueTrackerIssueThread(SimpleItem.SimpleItem, IssueTrackerIssue):
    """ Issuethreads class """

    meta_type = 'Issue Tracker Issue Thread'
    icon = '%s/issuethread.gif'%ICON_LOCATION
    
    meta_types = []

    _properties=({'id':'title',         'type': 'string', 'mode':'w'},
                 {'id':'comment',       'type': 'text',   'mode':'w'},
                 {'id':'threaddate',    'type': 'date',   'mode':'w'},
                 {'id':'fromname',      'type': 'string', 'mode':'w'},
                 {'id':'email',         'type': 'string', 'mode':'w'},
                 {'id':'display_format','type': 'string', 'mode':'w'},
                 )
    
    security=ClassSecurityInfo()

    manage_options = (
        {'label':'Properties', 'action':'manage_propertiesForm'}
        )

    def __init__(self, id, title, comment, threaddate, fromname, email,
                 display_format=None, REQUEST=None):
        """ create thread """
        self.id = str(id)
        self.title = title
        self.comment = comment
        if type(threaddate) == type('s'):
            threaddate = DateTime(threaddate)
        self.threaddate = threaddate
        self.fromname = fromname
        self.email = email
        if display_format:
            self.display_format = display_format
        else:
            self.display_format = self.default_display_format

            
    def index_html(self, REQUEST):
        " "
        return self.ShowIssueThread()

InitializeClass(IssueTrackerIssueThread)


############### ISSUETRACKERNOTIFYABLE ###################

class IssueTrackerNotifyable(SimpleItem.SimpleItem,
                             IssueTracker
                             ):
    """ IssueTrackerNotifyable class """

    meta_type = 'Issue Tracker Notifyable'
    icon = '%s/issuetracker_notifyable.gif'%ICON_LOCATION
    
    meta_types = []

    _properties=({'id':'alias',         'type': 'string', 'mode':'w'},
                 {'id':'email',         'type': 'string', 'mode':'w'},
                 {'id':'groups',        'type': 'lines',  'mode':'w'},
                 )
    
    security=ClassSecurityInfo()

    manage_editNotifyableForm = DTMLFile('dtml/editNotifyableForm', globals())
    
    manage_options = (
        {'label':'Properties', 'action':'manage_editNotifyableForm'},
        )
    
    def __init__(self, id, alias, email, groups=[]):
        """ init """
        if not ValidEmailAddress(email):
            raise "InvalidEmailAddress",\
                  "The email address (%s) was incorrect"%email
        else:
            self.id = id
            self.alias = alias.strip()
            self.email = email.strip()
            self.groups = groups

    def getName(self):
        """ Return self.alias. This might be in the future:
            self.firstname + self.lastname
        """
        return self.alias

    def manage_editNotifyable(self, alias=None, email=None, groups=None,
                              REQUEST=None):
        """ save changes to Notifyable """
        no = self
        n={'id':no.id}
        if alias is not None:   n['alias'] = alias.strip()
        else:                   n['alias'] = no.alias
        if email is not None:   n['email'] = email.strip()
        else:                   n['email'] = no.email
        if groups is not None:
            if type(groups)!=type([]):
                groups = [groups]
            n['groups'] = groups
        else:
            n['groups'] = no.groups
        n['groups'] = self._filterNotifyGroups(n['groups'])
            
        self._saveNotifyables([n])
        return self.manage_editNotifyableForm(REQUEST,
                        manage_tabs_message='Notifyable updated.')

    def getEmail(self):
        """ Return self.email """
        return self.email

##    def fixGroups(self, ok_groups):
##        """ make sure this notifyable doesn't belong to anything
##            not in ok_groups """
        

        
InitializeClass(IssueTrackerNotifyable)


############### ISSUETRACKERNOTIFICATION ###################

class IssueTrackerNotification(SimpleItem.SimpleItem,
                               IssueTracker
                               ):
    """ Issue Tracker Notification """

    meta_type='Issue Tracker Notification'
    icon = '%s/notification.gif'%ICON_LOCATION

    security=ClassSecurityInfo()
    
    _properties=({'id':'title',       'type': 'string',  'mode':'w'},
                 {'id':'change',      'type': 'string',  'mode':'w'},
                 {'id':'issueID',     'type': 'string',  'mode':'w'},
                 {'id':'comment',     'type': 'text',    'mode':'w'},
                 {'id':'emails',      'type': 'lines',   'mode':'w'},
                 {'id':'anchorname',  'type': 'string',  'mode':'w'},
                 {'id':'fromname',    'type': 'string',  'mode':'w'},
                 {'id':'date',        'type': 'date',    'mode':'w'},
                 {'id':'dispatched',  'type': 'boolean', 'mode':'w'},
                 )

    def __init__(self, id, title, change, issueID, comment, emails,
                 anchorname, fromname, date=None, dispatched=0, REQUEST=None):
        """ create notification """
        issueID = str(issueID)
        self.id = str(id)
        self.title = title
        self.change = change
        self.issueID = issueID
        self.comment = comment
        self.emails = emails
        self.anchorname = anchorname
        self.fromname = fromname
        self.dispatched = dispatched
        if date is None:
            self.date = getNowdate()
        else:
            self.date = date

    def dispatch(self):
        """ dispatch self """
        self.REQUEST.set('loop_over',self)
        getattr(self, 'dispatcher.html')
        

InitializeClass(IssueTrackerNotification)

