# Peter Bengtsson <mail@peterbe.com>
# Utility scripts for IssueTrackerProduct
#

# python 
import string, sys
import re
from random import shuffle

# ZOPE DEVELOPMENT
sys.path.insert(0,'D:\\Zope\\Zope243\\lib\\python')
    
# zope
from DateTime import DateTime
from DocumentTemplate import sequence
from Products.PythonScripts.standard import html_quote, newline_to_br, \
                     structured_text, url_quote

# Language constants
HOUR = 'hour'
HOURS = 'hours'
YEAR = 'year'
YEARS = 'years'
MONTH = 'month'
MONTHS = 'months'
DAY = 'day'
DAYS = 'days'
AND = 'and'
FINISH = 'old'


def preParseEmailString(es, sep='; ', names2emails={}):
    """ Take any string and strip out only a string
        of email addresses delimited by 'sep'
        "Peter <mail@peterbe.com>, John;peter@grenna.net, Joe"
        =>
        "mail@peterbe.com; peter@grenna.net"
        But suppose names2email={'joe','joey@host.com')
        then you would expect:
        "mail@peterbe.com; peter@grenna.net; joey@host.com"
    """
    real_emails=[]
    es = es.replace(';',' ').replace(',',' ').replace('>',' ').replace('<',' ')
    transtab = string.maketrans('/ ','  ')
    es = string.translate(es,transtab,'?&!()<=>*#[]{}')
    if names2emails:
        for k,v in names2emails.items():
            es = es.lower().replace(k.lower(), v)
    for e in es.split():
        if '@' in e and ValidEmailAddress(e) and e not in real_emails:
            real_emails.append(e)
    if real_emails:
        return sep.join(real_emails)
    else:
        return None
        
def getFromstring(name, email, email_html=1, email_must=0, email_pure=0):
    """
    First we filter which combs to care about;
    then we filter them.
    """
    name_html_quoted =html_quote(name)
    email_html_quoted=html_quote(email)
    
    passit = 0
       
    # first filter
    if name <> '' or email <> '':
        if email_must:
            if ValidEmailAddress(email):
                passit = 1
            else:
                passit = 0
        else:
            passit = 1
    
    # then display
    if not passit:
        return 0
    else:
        if email_pure and ValidEmailAddress(email):
            return email
        elif email == '':
            return name
        else:
            if email_html:
                return string.strip('%s &lt;<a href="mailto:%s">%s</a>&gt;'%\
                                    (name_html_quoted, email_html_quoted, email_html_quoted))
            else:
                return string.strip('%s <%s>'%(name, email))
    return 0

def AddParam2URL(url, params={}):
    """ return url and append params but be aware of existing params """
    p='?'
    if p in url:
        p = '&'
    url = url + p
    for key, value in params.items():
        url = url + '%s=%s&'%(key, url_quote(value))
    return url[:-1]

def getLaterDate(days=365):
    """ return a dateobject in future in rfc822 format """
    later = getNowdate()+days
    return later.rfc822()

def fixDictofLists(dict):
    " throw it a dictionary and it returns the values lowercased "
    for key,value in dict.items():
        if type(value)==type([]):
            dict[key] = _lowercaseList(value)
        elif type(value)== type('s'):
            dict[key] = value.lower()
    return dict

def _lowercaseList(lst):
    " lowercase and strip all items in a list "
    nlst=[]
    for l in lst:
        nlst.append(l.strip().lower())
    return nlst

def getRandomString(length=10, loweronly=1):
    """ return a very random string """
    lowercase = 'abcdefghijklmnopqrstuvwxyz'+'0123456789'
    l = list(lowercase + lowercase.upper())
    shuffle(l)
    s = string.join(l,'')
    if len(s) < length:
        s = s + getRandomString(loweronly=1)
    s = s[:length]
    if loweronly:
        return s.lower()
    else:
        return s
    
    

def GenerateNotifyableId(email, length=10, int_length=None):
    """ generate a random id for a notifyable string """
    timestamp = str(int(DateTime()))
    email= email.replace('@','_at_')
    if int_length is not None:
        timestamp = timestamp[-int_length:]
    return email[:length]+timestamp

def timeSince(container, firstdate, seconddate):
    """
     Use two date objects to return in plain english the difference between them.
     E.g. "3 years and 2 days"
      or  "1 year and 3 months and 1 day"
     If less than 1 day, return number of hours.

     If there is "no difference" between them, return false.
    """

    fdo = firstdate
    sdo = seconddate

    day_difference = int(abs(sdo-fdo))

    years = day_difference/365
    months = (day_difference % 365)/30
    days = (day_difference % 365) % 30

    if days == 0:
        # use hours
        hours=int(abs(sdo-fdo)*24)
        if hours == 1:
            return "1 %s %s" % (HOUR, FINISH)
        elif hours > 0:
            return "%s %s %s" % (hours, HOURS, FINISH)
        else:
            # if the difference is smaller than 1 hour, 
            # return it false
            return 0
    else:
        s = []
        if years == 1:
            s.append('1 %s'%(YEAR))
        elif years > 1:
            s.append('%s %s'%(years,YEARS))
        
        if months == 1:
            s.append('1 %s'%MONTH)
        elif months > 1:
            s.append('%s %s'%(months,MONTHS))
        
        if days == 1:
            s.append('1 %s'%DAY)
        elif days > 1:
            s.append('%s %s'%(days,DAYS))
        
        if len(s)>1:
            return "%s %s" % (string.join(s,' %s '%AND), FINISH)
        else:
            return "%s %s" % (s[0], FINISH)


def TranslateStatus(container,orginal):
    # 1. parse underscore_properties into dictionaries.
    code2lang = {}
    lang2code = {}
    for line in container.statuses_translation_verbs:
        splitted = line.split('|')
        code = splitted[0].strip()
        lang = splitted[1].strip()
        code2lang[code]=lang
        lang2code[lang]=code

    if lang2code.has_key(original):
        return lang2code(original)
    elif code2lang.has_key(original):
        return code2lang(original)
    else:
        return original



def DateStr2DateDate(str):
    """ convert a date as a string to a proper date """
    return DateTime(str)

def cookIdAndTitle(s):
    """ if s='image1~Image One.gif' then return ['image1.gif','Image One']

        Testwork:
        s='image1~Image One.gif'    => ['image1.gif','Image One']
        s='image1.gif'              => ['image1.gif','']
        s='image1~.gif'             => ['image1.gif','']
        s='image1~Image One.gif.gif'=> ['image1.gif','Image One.gif']
    """
    if s.find('~') == -1:
        return s, ''
    splitted = s.split('~',1)
    id = splitted[0]
    rest = s.replace(id+'~','')
    if rest.rfind('.') == -1:
        return id, rest
    else:
        pos_last_dot = rest.rfind('.')
        ext = rest[pos_last_dot:]
        id = id.strip() + ext.strip()
        title = rest[0:pos_last_dot].strip()
        return id, title

def LineIndent(text, indent):
    """ indent each new line with 'indent' """
    text = indent+text
    text = text.replace('\n','\n%s'%indent)
    return text


def ShowDescription(text, display_format=''):
    """
    Display text, using harmless HTML
    """
    if display_format == 'structuredtext':
        st=text

        for k,v in {'<':'&lt;', '>':'&gt;'}.items():
            st = st.replace(k,v)

        st = structured_text(st)

        for k,v in {'&amp;lt;':'&lt;', '&amp;gt;':'&gt;'}.items():
            st = st.replace(k,v)
            
        # BUG in structured_text in Zope 2.4.0
        # it returns these annoying tags.
        for tag in ['<html>','<body>','</body>','</html>']:
            st = st.replace(tag, '')

        return st
    
    else:
        t = html_quote(text)
        t = newline_to_br(t)
        return t
    

def getNowdate():
    """
    This method determines the format 
    with which new objects get a date property
    """
    return DateTime()


def _ShouldBeNone(result): return result is not None
def _ShouldNotBeNone(result): return result is None

tests = (
    # Thank you Bruce Eckels for these
  (re.compile("^[0-9a-zA-Z\.\-\_]+\@[0-9a-zA-Z\.\-]+$"), _ShouldNotBeNone, "Failed a"),
  (re.compile("^[^0-9a-zA-Z]|[^0-9a-zA-Z]$"), _ShouldBeNone, "Failed b"),
  (re.compile("([0-9a-zA-Z]{1})\@."), _ShouldNotBeNone, "Failed c"),
  (re.compile(".\@([0-9a-zA-Z]{1})"), _ShouldNotBeNone, "Failed d"),
  (re.compile(".\.\-.|.\-\..|.\.\..|.\-\-."), _ShouldBeNone, "Failed e"),
  (re.compile(".\.\_.|.\-\_.|.\_\..|.\_\-.|.\_\_."), _ShouldBeNone, "Failed f"),
  (re.compile(".\.([a-zA-Z]{2,3})$|.\.([a-zA-Z]{2,4})$"), _ShouldNotBeNone, "Failed g"),
)
def ValidEmailAddress(address, debug=None):
    for test in tests:
        if test[1](test[0].search(address)):
            if debug: return test[2]
            return 0
    return 1

def ss(s):
    """ simple string """
    return s.strip().lower()



def test():
    print " ++ cookIdAndTitle() ++ "
    print cookIdAndTitle('image1~Image One.gif'), " AND ['image1.gif','Image One']"
    print cookIdAndTitle('image1.gif'), " AND ['image1.gif','']"
    print cookIdAndTitle('image1~.gif'), " AND ['image1.gif','']"
    print cookIdAndTitle('image1~Image One.gif.gif'), " AND ['image1.gif','Image One.gif']"
    print cookIdAndTitle('monk ~ Monkey.dtml'), " AND ['monk.dtml','Monkey']"

    print " ++ ValidEmailAddress() ++ "
    print ValidEmailAddress(''), " AND 0"
    print ValidEmailAddress('issuetracker@peterbe.com'), " AND 1"
    print ValidEmailAddress('issuetracker @peterbe.com'), " AND 0"

    print " ++ getRandomString() ++ "
    print getRandomString(), " "
    print getRandomString(length=50), " AND (length=50)"
    print getRandomString(loweronly=1), " AND (loweronly=1)"
    
if __name__=='__main__':
    test()

##### HMMMM #############

def getMailHost(self,id='MailHost'):
     """ since unrestricted aquisition to the MailHost (or whatever id)
         object is impossible, we do it via a External Method instead.
     """
     return getattr(self, id)
			    

##def sendIssue2Friends(self, issueID, email_string, ignoreword=''):
##    """ 
##    use an email_string and a predefined message to send out
##    information to people about a new issue.
##    """
##    request = self.REQUEST
##    mailhost = self.MailHost
##    
##    issue = getattr(self, issueID)
##    
##    if email_string.strip() == '' or email_string.strip() == ignoreword:
##	request.RESPONSE.redirect(issue.absolute_url())
##	
##    br = "\r\n"
##    
##    fromname = None
##    fromemail = None
##    
##    if self.has_cookie(self.name_cookiekey):
##	fromname = self.get_cookie(self.name_cookiekey)
##    if self.has_cookie(self.email_cookiekey):
##	fromemail = self.get_cookie(self.email_cookiekey)
##	
##    if fromname is None and fromemail is not None:
##	fr = fromemail
##    elif fromname is not None and fromemail is not None:
##	fr = "%s <%s>"%(fromname, fromemail)
##    else:
##	fr = "%s <%s>"%(self.sitemaster_name, self.sitemaster_email)
##	
##    to = email_string
##    
##    subject = "An Issue added to your attention"
##    
##    if fromname is None:
##	msg = "An issue has been added to your attention at \
##	%s with the following title:"%(self.title) + br
##    else:
##	msg = fromname + " has entered an Issue in the %s and \
##	wanted you to know about it, with the following title:"%(self.title) + br
##		
##	msg = msg + ' "' + issue.title + '" ' + 2*br
##	msg = msg + "The Issue can be found at" + br
##	msg = msg + self.ActionURL(url=issue.absolute_url()) + br*3     
##	msg = msg + "Regards, " + self.sitemaster_name
##	
##    body = string.join(("From: %s"%fr,
##                 "To: %s"%to,
##		 "Subject: %s"%subject,
##		 "",msg),br)
##		 
##    try:
##	mailhost.send(body, to, fr, subject)
##	qs = "?NewIssue=Send2Friends+Success"
##    except:
##	qs = "?NewIssue=Send2Friends+Failed"
##
##    request.RESPONSE.redirect(issue.absolute_url()+qs)

def unindex_object(self):
    """A common method to allow Findables to unindex themselves."""
    self.aq_parent.uncatalog_object('/'.join(self.getPhysicalPath()))
	
def hasStoredFilter(self):
    """
        Return true if cookieobject has a xxx_filteroptions
        variable stored.
    """
    trackerroot = getRoot(self)

    tfc = "%s_filteroptions"%trackerroot.id
    
    if self.has_cookie(tfc):
        return 1
    else:
        return 0

    
def FillListingRequest(self):
    """
        Prepopulate REQUEST
        with saved filteroptions from cookie
        Don't use this if the url is fiddled with manually,
        like in url?type=bug
    """
    request = self.REQUEST
    trackerroot = getRoot(self)

    #TrackerFilteroptionsCookiename
    tfc = "%s_filteroptions"%trackerroot.id
    
    usecookie=1
    # things expected to be set manually in URL
    reasons_toignore= ['section','type','urgency','status','fromname','email']
    for reason in reasons_toignore:
        if reason in request.form.keys():
            usecookie = 0

    if usecookie and request.has_key('reset_filter') and self.has_cookie(tfc):
        # remove cookie and move on
        self.delete_cookie(tfc)

    if usecookie and self.has_cookie(tfc) and not request.has_key('filteroptions'):
        for key,value in _DictStr2DictDict(self.get_cookie(tfc)).items():
            request.set(key,value)
            
    return "Filled Listing Request"

def FillRequest(self):
    """
    Loop though a list of variables
    if it doesn't exist in either self
    or REQUEST, fill them into REQUEST
    as '' values
    """
    request = self.REQUEST

    propertiesNulls 	= ['subject','fromname','email','description',
                           'url2issue','urgency','type','display_format']
    propertiesZero 	= ['confidential','hide_me']
    propertiesEmpty 	= ['sections']
    
    nullified = []
    for property in propertiesNulls:
        if not request.has_key(property):
            request.set(property,'')
            
    for property in propertiesZero:
        if not request.has_key(property):
            request.set(property,0)
            
    for property in propertiesEmpty:
        if not request.has_key(property):
            request.set(property,[])
    
    # Also, for all list variables, make sure that only
    # one item doesn't make it a string
    for property in propertiesEmpty:
        if type(request[property]) == type(''):
            request[property] = [request[property]]
            
    # properties that have default values
    propertiesDefault	= [['status','open']]
    for property in propertiesDefault:
        if not request.has_key(property[0]):
            request.set(property[0],property[1])

    # special one
    try:
        if request['url2issue'][0] != '/' and request['url2issue'][:4] !='http':
            request.set('url2issue','http://%s'%request['url2issue'])
    except:pass # no big deal


    return ""
    

    
def SubmitIssue(self):
    """
    1) Check data
    2) Try to create issue
    	2a) If success, RESPONSE.redirect to issue plus Thank you message
    	2b) If failure, print failed data and urge to submit again
    """
    req = self.REQUEST
    
    randomID_int_length = 5
    if hasattr(self, 'randomID_int_length'):
        # use the Zope version instead
        randomID_int_length = self.randomID_int_length

    failures = {}
    has_cookie = self.has_cookie
    get_cookie = self.get_cookie
    set_cookie = self.set_cookie
    
    #
    # Tune data
    #
    
    # strip whitespace
    for property in ['subject','fromname','email','url2issue','display_format']:
        request[property] = request[property].strip()

    # quick add doesn't necessarily ask for email,
    # when it already exists in the cookie. (lets do it for name too)
    if request.email=='' and has_cookie(self.email_cookiekey):
        request.set('email',get_cookie(self.email_cookiekey))
    if request.fromname=='' and has_cookie(self.name_cookiekey):
        request.set('fromname',get_cookie(self.name_cookiekey))
        
        
        
    # set sections to default values if empty
    if len(request.sections) ==0:
        request.sections = self.defaultsections
    
    #
    # Check data
    #
    
    # subject mustnt be empty
    if request.subject =='':
        # no input
        failures['Subject'] = "You must at least enter a subject"
        
    # to prevent problems with sending mail
    if not ValidEmailAddress(request.email):
        failures['Invalid Emailaddress'] = "To be able to notify you about the Issue, you must enter a valid email address."
        # hack. Leave a trace in the REQUEST
        # for the QuickAddIssue
        request.set('invalidemailaddress',1)
    
    if len(failures) >0:
        return self.ExplainSubmitErrors(failures)
        
    else:
        #
        # Let's Add it!!
        #
        
        # 1) Generate Random Issue Id
        if hasattr(self,'issueprefix'):
            prefix = self.issueprefix
        else:
            prefix = ""

        genid = generateID(self,"Issue", 5,prefix)
        
        # 3) Set cookie of name and email using cookieobject
        if request.has_key('fromname') and request.fromname <> '':
            set_cookie(self.name_cookiekey, request.fromname)
        if request.has_key('email') and request.email <> '':
            set_cookie(self.email_cookiekey, request.email)
        if request.has_key('display_format') and \
           request.display_format in self.display_formats:
            set_cookie(self.display_format_cookiekey,request.display_format)
            
        # 4) Try to add Issue
        # add Issue
        self.IssueTrackerIssue.createInObjectManager(genid,self.REQUEST)
                
        # get object from id
        issueobject = getattr(self, genid)

        # prepopulate empty REQUEST data
        FillRequest(self)

        # filter out empty item from sections
        sections_newlist=[]
        for section in request.sections:
            if section!='':
                sections_newlist.append(section)
                
        # add all the properties
        props = {}
        props['title'   ]       = request.subject
        props['fromname']	= request.fromname
        props['email'   ]	= request.email
        props['url2issue'] 	= request.url2issue
        props['type'    ] 	= request.type
        props['urgency' ] 	= request.urgency
        props['description'] 	= request.description
        props['display_format'] = request.display_format
        props['confidential'] 	= int(request.confidential)
        props['hide_me' ]	= int(request.hide_me)
        props['issuedate'] 	= getNowdate()
        props['status'  ]	= request.status
        props['sections']       = sections_newlist

        # update issue object
        issueobject.propertysheets.properties.manage_changeProperties(props)
        
        # Catalog yourself
        self.catalog_object(issueobject)

        # 5) Redirect
        redirect_url = '%s?NewIssue=Submitted'%(issueobject.absolute_url())
        request.RESPONSE.redirect(redirect_url)
        
        # Still here?
        # Hmm show a simple URL
        # Remember that Zope expects this method to return
        # a list of failures.
        return ['<h2><a href="%s">Issue Added</h2>'%redirect_url]
        

def generateID(catalog, start, length, prefix='', meta_type='Issue Tracker Issue',increment=None):
    # digits
    if increment is None:
        idnr=len(catalog.objectIds(meta_type))+1
        increment = idnr+1
    else:
        increment=increment+1
        idnr=increment
        
    r=string.zfill(str(idnr),length)
    id = "%s-%s%s"%(start, prefix, r)
    if hasattr(catalog, id):
        # ah! Id already exists, try again
        return generateID(catalog, start, length,meta_type,increment=increment)
    else:
        return id




def getRoot(self):
    """ superValues() does what objectValues() does, but
    continues down the path until found. If two Issue Tracker
    instances are in the same folder, they are picked out by
    id unsmartly, hence this little test"""
    supervalues = self.superValues('Issue Tracker')
    for supervalue in supervalues:
        if supervalue.id in self.REQUEST.URL1.split('/'):
            return supervalue
    else:
        return supervalues[0]
    
def unihack(s):
    """ unicode hack """
    return s
   # return string.join(list(s),'')


def ListIssueFiltered(self):
    """ wrapper around _ListIssueFiltered() that prepares a search
    if REQUEST holds 'q'
    """
    if self.REQUEST.has_key('q') and self.REQUEST['q'] != '':
        q=self.REQUEST['q']
	q=q.replace('%','*') # allow both wildcards
        catalogs = self({'title':q}) + \
                   self({'description':q}) + \
                   self({'fromname':q}) + \
                   self({'email':q})
        seq=[]
        for brain in catalogs:
            try:
                issueobject = getattr(self,brain.id)
                if issueobject not in seq:
                    seq.append(issueobject)
            except:
                # probably because issue has been deleted manually
                pass
    else:
        seq = self.objectValues('Issue Tracker Issue')
    return _ListIssueFiltered(self,seq)

def _ListIssueFiltered(self,objsequence):
    rlist = []
    req = self.REQUEST

    set_cookie = self.set_cookie

    try:
        reverse=int(request.reverse)
    except:
        reverse=0

    trackerroot = getRoot(self)

    #TrackerFilteroptionsCookiename
    tfc = "%s_filteroptions"%trackerroot.id
    
    # remember filteroptions
    if request.has_key('filteroptions'):
        tracker_filteroptions={}
        for each in ['showtypes','showsections','showstatuses','showfromname','showurgencies','showemail']:
            try:
                tracker_filteroptions[each]=request[each]
            except:pass
        set_cookie(tfc,tracker_filteroptions)
        
   
    #
    # prepare a master filter
    #
    
    filter ={}
    
    # statuses
    if request.has_key('showstatuses'):
        filter['status'] = request.showstatuses
    elif request.has_key('status'):
        filter['status'] = [request.status]
    else:
        filter['status'] = self.statuses
        
    # sections
    if request.has_key('showsections'):
        filter['sections'] = request.showsections
    elif request.has_key('section'):
        filter['sections'] = [request.section]
    else:
        filter['sections'] = self.sections_options

    # urgencies
    if request.has_key('showurgencies'):
        filter['urgency'] = request.showurgencies
    elif request.has_key('urgency'):
        filter['urgency'] = [request.urgency]
    else:
        filter['urgency'] = self.urgencies
        
    # types
    if request.has_key('showtypes'):
        filter['type'] = request.showtypes
    elif request.has_key('type'):
        filter['type'] = [request.type]
    else:
        filter['type'] = self.types
        
    # fromname
    if request.has_key('showfromname') and \
       request.showfromname <> self.when_ignore_word:
            filter['fromname'] = string.strip(request.showfromname)
    elif request.has_key('fromname'):
        filter['fromname'] = request.fromname
    else:
        filter['fromname'] = ''

    # email
    if request.has_key('showemail') and \
       request.showemail <> self.when_ignore_word:
            filter['email'] = string.strip(request.showemail)
    elif request.has_key('email'):
        filter['email'] = request.email
    else:
        filter['email'] = ''


    # prepare for caseindependence
    filter = _fixDictofLists(filter)
        
    # all Issue objects
    containers = objsequence 

    #
    # Before we can determine what to sort on
    # we need to check the name of the properties
    
    # the default
    sortorder = ('bobobase_modification_time',)
    skipsort = 0
    if request.has_key('sortorder') and request['sortorder'] != 'date':
        sortorder = (request['sortorder'],)
        if request['sortorder'] == 'from':
            sortorder = ('fromname',)
        if request['sortorder'] == 'creationdate':
            sortorder = ('issuedate',)
        if request['sortorder'] == 'changedate':
            sortorder = ('bobobase_modification_time',)

        if request['sortorder'] == 'status':
            #
            # Very special!
            # self.statuses is an "human ordered" list like 'open','taken',...'completed','forgotten'
            statuses = {}
            # create a dictionary of list divided in groups by their status
            # like this: {'open':['foo','bar'], ... 'forgotten':['this','that']}
            for con in containers:
                if statuses.has_key(con.status):
                    statuses[con.status].append(con)
                else:
                    statuses[con.status] = [con]
            # recreate the containers
            containers = []
            # remembers that self.statuses is ordered
            for status in self.statuses:
                # perhaps there are no 'forgotten' issues even
                # though there is 'forgotten' in self.statuses
                if statuses.has_key(status):
                    # order all issues with this status by bobobase_modification_time
                    thisstatus_containers = sequence.sort(statuses[status],(('bobobase_modification_time',),))
                    # reverse bobobase_modification_time sort naturally
                    thisstatus_containers.reverse()
                    containers = containers + thisstatus_containers
                    # skip the default sort later
                    skipsort = 1
                    
        if request['sortorder'] == 'urgency':
            #
            # Very special too
            # read the comments on "if request['sortorder'] == 'status':" above
            urgencies = {}
            for con in containers:
                if urgencies.has_key(con.urgency):
                    urgencies[con.urgency].append(con)
                else:
                    urgencies[con.urgency] = [con]
            containers = []
            for urgency in self.urgencies+['']:
                if urgencies.has_key(urgency):
                    thisurgency_containers = sequence.sort(urgencies[urgency], (('bobobase_modification_time',),))
                    thisurgency_containers.reverse()
                    containers = containers + thisurgency_containers
                    skipsort = 1
            
    else:
        # if no sortorder,
        # it uses the bobobase_modification_time
        # and this has to be reversed so that
        # the latest is listed first
        reverse =1

    # do the basic sort.
    # 'sortorder' is a tuple
    if not skipsort:
        containers = sequence.sort(containers,(sortorder,))

        
    if reverse:
        containers.reverse()

    # containers is not order, just filter it
    for con in containers:
        # if confidential and (if) Manager show, if not skip
        if not con.confidential or self.hasManagerRole():
            if filter == {}:
                rlist.append(con)
            else:
                passsections = 0
                for section in con.sections:
                    if ss(section) in filter['sections']:
                        passsections = 1
                        break
                # pass the filter lists

		if ss(con.status) in filter['status'] \
                   and ss(con.urgency) in filter['urgency'] \
                   and ss(con.type) in filter['type'] \
                   and passsections:


                    if (ss(con.fromname) == filter['fromname'] or filter['fromname']=='') \
                       and \
                       (ss(con.email) == filter['email'] or filter['email'] =='' ):
                        rlist.append(con)
	    
    return rlist

           

def _fixDictofLists(dict):
    " throw it a dictionary and it returns the values lowercased "
    for key,value in dict.items():
        if type(value)==type([]):
            dict[key] = _lowercaseList(value)
        elif type(value)== type('s'):
            dict[key] = value.lower()
    return dict

def _lowercaseList(lst):
    " lowercase and strip all items in a list "
    nlst=[]
    for l in lst:
        nlst.append(l.strip().lower())
    return nlst

            
                
def ModifyIssue(self,action='Addfollowup'):
    req = self.REQUEST

    
    action = action.strip().lower()
    oldstatus = self.status

    if hasattr(self,'issueprefix'):
        prefix = self.issueprefix
    else:
        prefix = ''
            
    if not request.has_key('comment_description'):
        request.set('comment_description','')

    if not request.has_key('display_format'):
        request.set('display_format',self.default_display_format)

    req_manager = 1 # require manager
    dummychange = 0
    addfollowup = 0
    past_tense=None
    if action == 'open':
        past_tense='open'
        addfollowup = 1
    if action == 'take':
        past_tense='taken'
        addfollowup = 1
    if action == 'put on hold':
        past_tense='on hold'
        addfollowup = 1
    if action == 'reject':
        past_tense='rejected'
        addfollowup = 1
    if action == 'complete':
        past_tense='completed'
        addfollowup = 1
    if action == 'delete':
        DeleteIssue(self, self.id)
        addfollowup = 0
        redirect_url = request.URL2
    if action == 'addfollowup':
        addfollowup = 1
        dummychange = 1
        req_manager = 0

    if req_manager and not self.hasManagerRole():
        # the chosen action requires manager role,
        # but the person is not a manager
        getattr(self,'login')(self,req)

    if dummychange:
        self.propertysheets.properties.manage_changeProperties(\
            {'status':self.status})
        
    if past_tense is not None:
        self.propertysheets.properties.manage_changeProperties(\
            {'status':past_tense})
    else:
        # anything else means that the comment_description
        # cannot be empty
        if request['comment_description'].strip()=='':
            addfollowup=0
            redirect_url = request.URL1
    
    # most actions may perhaps add a little comment
    if addfollowup:

        genid = generateID(self, "IssueThread", self.randomID_int_length, \
                           prefix, meta_type='Issue Tracker Issue Thread')
        
        if action == 'addfollowup':
            gentitle = 'Added Issue followup'
        else:
            gentitle = 'Changed status from %s to %s'%\
                       (oldstatus.capitalize(), past_tense.capitalize())

        # create an Thread object
        self.IssueTrackerIssueThread.createInObjectManager(genid,req)

        # get that object
        followupobject = getattr(self, genid)

        # fix variables
        try:
            fromname = request.fromname
        except:
            fromname = ''
        try:
            email = request.email
        except:
            email = ''
        try:
            display_format = request.display_format
        except:
            display_format = self.default_display_format

            
        # remember visitor
        self.set_cookie(self.name_cookiekey,fromname)
        self.set_cookie(self.email_cookiekey,email)
        self.set_cookie(self.display_format_cookiekey,display_format)

        # update Thread object
        newproperties = {'title':gentitle,
                         'comment':request.comment_description,
                         'threaddate':getNowdate(),
                         'fromname':fromname,
                         'email':email,
                         'display_format':display_format}
        followupobject.propertysheets.properties.manage_changeProperties(newproperties)

        if request.has_key('notify'):
                    
            # now, create and email-alert-queue object
            # using filtered email address
            # get who to notify        
            email_addresses = Others2Notify(self, genid, do='email',emailtoskip=email)

            if len(email_addresses) > 0:

                # get root instance
                trackerroot = getRoot(self)
                notifyid = generateID(trackerroot, 'Notification',5, prefix, meta_type='Issue Notification')
                trackerroot.IssueNotification.createInObjectManager(notifyid,req)
                notifyobject = getattr(trackerroot,notifyid)
                newproperties = {'title':self.title,
                                 'change':gentitle,
                                 'issueID':self.id,
                                 'comment':request.comment_description,
                                 'anchorname':len(self.objectIds('Issue Tracker Issue Thread')),
                                 'emails':email_addresses,
                                 'date':getNowdate(),
                                 'fromname':fromname}
                notifyobject.propertysheets.properties.manage_changeProperties(newproperties)
                
                # use the dispatcher to try to send
                # this notification right now.
                request.set('loop_over',[notifyobject])
                # there is no big deal if the dispatcher crashes here
                # because the notification is saved and the dispatcher
                # can be invoked some other time manually
                try:
                    getattr(self,'dispatcher.html')(self,req)
                except: pass 
        
        redirect_url = '%s#%s'%\
                       (self.absolute_url(),len(self.objectIds('Issue Tracker Issue Thread')))

    # ready ! redirect!
    request.RESPONSE.redirect(redirect_url)
            
    
def Others2Notify(self, issueID, do='print',emailtoskip=None):
    """
    Returns a list of names and emails of guys to notify on this Issue. 
    Take the hide_me and confidential factor into consideration.
    
    if do == 'print':
        display the people on screen (using hide_me and confidential factor)
    if do == 'email':
        return a pure list of all email address
    """

    if emailtoskip is None:
        if self.has_cookie(self.email_cookiekey):
            emailtoskip = self.get_cookie(self.email_cookiekey)
        else:
            emailtoskip = '@@@@@'

    fromstrings = []


    issue = self

    # proceed only if Manager or open-name on the issue
    if self.hasManagerRole() or not issue.hide_me:
        # self.email is the email of the thread
        # emailtoskip comes from the param/cookie if applicable
        if not issue.email == emailtoskip:
            if do =='email':
                fromstring = _fromstring(issue.fromname, issue.email, email_must=1, email_pure=1)
            else:
                fromstring = _fromstring(issue.fromname, issue.email)
            # fromstring looks like this now: "Peter Bengtsson <mail@peterbe.com>"
            # or ['mail@peterbe.com'] depending on 'do'
            # ...maybe
            if fromstring and fromstring not in fromstrings:
                fromstrings.append(fromstring)

    for thread in issue.objectValues('Issue Tracker Issue Thread'):
        if not thread.email == emailtoskip:
            if do =='email':
                fromstring = _fromstring(thread.fromname, thread.email, email_must=1, email_pure=1)
            else:
                fromstring = _fromstring(thread.fromname, thread.email)
                
            if fromstring and fromstring not in fromstrings:
                
                    fromstrings.append(fromstring)

    return fromstrings


def _filterduplicates(olist):
    nlist = []
    flist = []
    for each in olist:
        if not each[0:1] in nlist:
            nlist.append(each[0:1])
            flist.append(each)
    return flist
    
    
    


def HeadingLinks(self, display, sortname, default=0, startreversed=0):
    """Returns a hyperlink that can
       be used for resorting the listing
    """
    req = self.REQUEST
    querystring = request.QUERY_STRING

    ascending = self.ascarrow.tag(hspace=2)
    descending = self.descarrow.tag(hspace=2)
    invisible = self.invisiblearrow.tag(hspace=2)
    
    
    n_querystring = []
    for query in string.split(querystring,'&'):
        if string.find(query,'sortorder') == -1 and string.find(query,'reverse') == -1:
            n_querystring.append(query)
    querystring = string.join(n_querystring,'&')

    if querystring != '':
        querystring = '&%s'%querystring
    
    if request.has_key('reverse') and (request['reverse'] == 1 or request['reverse'] == '1'):
        reverse = 1
    elif startreversed and not (request.has_key('sortorder') and request['sortorder']==sortname):
        reverse = 1
    else:
        reverse = 0
    if request.has_key('sortorder') and request['sortorder'] == sortname or (default and not request.has_key('sortorder')):
        # have sorted by this, just let them reverse
        if reverse:
            return '%s<a class="headinglink_current" href="%s?sortorder=%s%s" title="Click to sort by %s">%s</a>' \
                   % (descending,request.URL, sortname, querystring, display, display )
        else:
            return '%s<a class="headinglink_current" href="%s?sortorder=%s&reverse=1%s" title="Click one more time and the sorting will be reversed">%s</a>'\
                   % (ascending,request.URL, sortname, querystring, display )
    else:
        if startreversed:
            return '%s<a class="headinglink" href="%s?sortorder=%s%s&reverse=1" title="Click to sort by %s">%s</a>' \
               % (invisible,request.URL, sortname, querystring, display, display )
        else:
            return '%s<a class="headinglink" href="%s?sortorder=%s%s" title="Click to sort by %s">%s</a>' \
               % (invisible,request.URL, sortname, querystring, display, display )
        
def MarkNotificationDispatch(self):
    """ set the 'dispatch' unrestricted of security """
    self.propertysheets.properties.manage_changeProperties({'dispatched':1})

def ShowAboutElsewhere(self):
    """ return the About page of the product instance elsewhere unrestricted """
    return self.aq_parent.About(self,self.REQUEST)


def _ListStr2ListList(ListStr):
    """ Return ['a','b'] from "['a','b']" unsecurly """
    if type(ListStr) == type([]):return ListStr
    if type(ListStr) == type('s'):
        if ListStr[0] =='[' and ListStr[-1]==']':
            exec "AsList = " + ListStr
            return AsList
    else:
        raise "NotStringNorList", \
              "The input variable was not a list or a string as expected"

def _DictStr2DictDict(DictStr):
    """ Return {'a':'A'} from "{'a':'A'}" unsecurly """
    if type(DictStr) == type({}): return DictStr
    if type(DictStr) == type('s'):
        if DictStr[0]=='{' and DictStr[-1]=='}':
            exec "AsDict = " + DictStr
            return AsDict
        else:
            raise "NotDictionaryString", \
                  "The string does not look like a dictionary"
    else:
        raise "NotStringNorDict", \
              "The input variable was not a dictionary or a string as expected"
        
    
    
#---------------------------------------------------
#
# Other utils

def siteIssue2issueIssue(self, sitetracker, issuetracker):
    """ copy objects from a sitetracker instance to an issuetracker instance """
    if sitetracker.meta_type != 'Site Tracker' and \
       issuetracker.meta_type != 'Issue Tracker':
        raise "WrongTrackers", "Parameters should be (a_sitetracker_instance, an_issuetracker_instance)"

    res = ''
    context = self
    site = sitetracker
    issue = newissuetracker

    from time import sleep

    # transfer all issues from site to issue.

    res = "++ OLD ISSUES ++ \n"
    oldsiteIssues = site.objectValues('Site Tracker Issue')
    for o in oldsiteIssues:
        # create Issue Tracker Issue
        issue.IssueTrackerIssue.createInObjectManager(o.id,{})

        res += o.title + '\n'

        props = {}
        props['title'   ]       = o.title
        props['fromname']       = o.fromname
        props['email'   ]       = o.email
        props['url2issue']      = o.url2issue
        props['type'    ]       = o.type
        props['urgency' ]       = o.urgency
        props['description']    = o.description
        props['display_format'] = 'structuredtext'
        props['confidential']   = int(o.confidential)
        props['hide_me' ]       = int(o.hide_me)
        props['issuedate']      = o.issuedate
        props['status'  ]       = o.status
        props['sections']       = o.sections
        
        newissue = getattr(issue, o.id)
        newissue.propertysheets.properties.manage_changeProperties(props)
        issue.catalog_object(newissue)

        sleep(3)

        for f in o.objectValues('Site Tracker Issue Thread'):
            newissue.IssueTrackerIssueThread.createInObjectManager(f.id,{})
            
            sprops={}
            sprops['title'] = f.title
            sprops['comment']= f.comment
            sprops['threaddate'] = f.threaddate
            sprops['fromname'] = f.fromname
            sprops['email'] = f.email
            sprops['display_format'] = 'structuredtext'        
            
            newthread = getattr(newissue, f.id)
            newthread.propertysheets.properties.manage_changeProperties(sprops)
            
            sleep(2)


    res += "++ OLD NOTIFICATIONS ++\n"

    oldnotifications = site.objectValues('Issue Notification')
    for o in oldnotifications:
        # create Issue Notification
        issue.IssueNotification.createInObjectManager(o.id,{})

        res += o.title +'\n'

        props = {}
        props['title'      ] = o.title
        props['change'     ] = o.change
        props['issueID'    ] = o.issueID
        props['comment'    ] = o.comment
        props['emails'     ] = o.emails
        props['anchorname' ] = o.anchorname
        props['fromname'   ] = o.fromname
        props['date'       ] = o.date
        props['dispatched' ] = o.dispatched

        newnotification = getattr(newissue, o.id)
        newnotification.propertysheets.properties.manage_changeProperties(props)
     
        sleep(2)

    return res



def IssueTracker_add(self):
    """ add through a External Method instead """

    req = self.REQUEST
    if not request.has_key('title'):
        request.set('title', 'Issue Tracker')

    self = self.manage_addProduct['IssueTrackerProduct']
    # create IssueTracker
    
    new_inst = self.IssueTracker.createInObjectManager(request['id'], req)
    self.setZCatalogProperties(request['id'])

    # add standard objects
    standards = self.Control_Panel.Products.IssueTrackerProduct.standards
    for object in standards.objectValues():
        meta_type = object.meta_type
        id = object.getId()
        title = object.title
        file = getattr(standards, id)
        if meta_type == 'DTML Document':
            new_inst.manage_addDTMLDocument(id, title, file=file)
        elif meta_type == 'Image':
            if hasattr(object, 'belong2folder'):
                if not hasattr(new_inst, object.belong2folder):
                    new_inst.manage_addFolder(object.belong2folder)
                new_inst[object.belong2folder].manage_addImage(id, file=file.data, title=title)
            else:
                new_inst.manage_addImage(id, file=file.data, title=title)
            
                
            


    # whereto next?
    if request.has_key('addandedit'):
        request.RESPONSE.redirect(getattr(self, request['id']).absolute_url()+'/propertysheets/properties/manage')

    elif request.has_key('DestinationURL'):
        request.RESPONSE.redirect(request.DestinationURL+'/manage_workspace')

    else:
        request.RESPONSE.redirect(request.URL1+'/manage_workspace')

    
if __name__=='__main__':
    print "TEST getLaterDate()"
    print getLaterDate()
    print getLaterDate(14)

    print "----------\n"
    print "TEST AddParam2URL()"
    print AddParam2URL('http://www.peterbe.com')
    print AddParam2URL('http://www.peterbe.com',{'a':'A','b':'B'})
    message="""Line1
Line2"""
    print AddParam2URL('http://www.peterbe.com?o=O',{'a':message,'b':'B spaced'})

    print "----------\n"
    print "TEST preParseEmailString()"
    print preParseEmailString("mail@peterbe.com;JOE,peter@grenna.net,Peter Bengtsson<ppp@ppp.se>; Peter <pete@peterbe.com", \
                              names2emails={'Joe':'joey@peterbe.com', 'Peter bengtsson':'ppp@ppp.se'})
    print preParseEmailString("mail@ ", \
                              names2emails={'Joe':'joey@peterbe.com', 'Peter bengtsson':'ppp@ppp.se'})
    
    
